/*	MVEPlayer.c - implements MVEPlayer function.
	Author: Shaktazuki
	with generous help from Mike Melanson, of FFMpeg and www.multimedia.cx fame
	Major goals:
	TODO: implement all video decoder(s)
	TODO: fix BG movie compatibility
	Minor goals:
	TODO: dragging the video output around messes up the file reads; pause while dragging or disable dragging (SDL_NOFRAME)
		or ... the callback can get its own data from the file using the file's audio frame indices (probably a bad idea).
	TODO: generalize framerate timing fix
	TODO: account for truecolor flag in opcode 0x05 
	TODO: change flag checking from CHKFLAG(FLAG) to just FLAG and make unsetting something like ~FLAG
	(may need to do that before this:)
	TODO: implement non-stereo audio decoder
	TODO: redefine mve data structs and code that uses them.
	(with respect to this last: the more I think about it, the original MVE stuff is nearly ideal).
*/
#include "MVEPlayer.h"
#include <assert.h>
#ifdef _MSC_VER
#include "sdl.h"
#else
#include <SDL.H>
#include <SDL/SDL.H>
#endif

/* define (potentially) handled opcode (opcode) types */
#define STOP_PLAYBACK			0x00
#define FETCH_NEXT_CHUNK		0x01
#define CREATE_TIMER			0x02
#define INIT_AUDIO_BUFFERS		0x03
#define START_AUDIO				0x04
#define INIT_VIDEO_BUFFERS		0x05
#define SEND_BUFFER_TO_DISPLAY	0x07
#define AUDIO_FRAME				0x08
#define SET_PALETTE				0x0C
#define SET_DECODING_MAP		0x0F
#define VIDEO_DATA				0x11

/**************/
/* flag masks */
/**************/

/* flow control flags */
#define QUITTING			0x00000001
#define PAUSED				0x00000002

/* timer flags */
#define TIMER_CREATED		0x00000004
#define TIMER_INIT			0x00000008

/* audio flags */
#define AUDIO_PLAYING		0x00000010
#define AUDIO_COMPRESSED	0x00000020
#define AUDIO_STEREO		0x00000040
#define SAMPLESIZE16		0x00000080

/* flags declaration */
Uint32 flags = 0;

/**************************/
/* internally used macros */
/**************************/

#define SETFLAG(x) flags |= x
#define NEGFLAG(x) flags &= ~x
#define CHKFLAG(x) (flags & x)
/* fetch LE words from a bytestream x */
#define LE16(x) (*(x) | ((*(x+1)) << 8))
/* write a signed word y to bytestream x */
#define WRITEWORD(x,y) {Sint16 * temp = (Sint16 *)x; *temp = y;}

/******************************/
/* internally used structures */
/******************************/

typedef struct CHUNK
{
	Uint16 length;
	Uint16 type;
} chunk;

typedef struct OPCODE
{
	Uint16 length;
	Uint8 type;
	Uint8 version;
} opcode;

typedef struct BUFSTRUCT
{
	Uint32 length;
	Uint8 *data;
} buffer;

/* Interplay DPCM table */
Sint16 interplay_delta_table[256] =
{
         0,      1,      2,      3,      4,      5,      6,      7,      8,      9,     10,     11,     12,     13,     14,     15,
        16,     17,     18,     19,     20,     21,     22,     23,     24,     25,     26,     27,     28,     29,     30,     31,
        32,     33,     34,     35,     36,     37,     38,     39,     40,     41,     42,     43,     47,     51,     56,     61,
        66,     72,     79,     86,     94,    102,    112,    122,    133,    145,    158,    173,    189,    206,    225,    245,
       267,    292,    318,    348,    379,    414,    452,    493,    538,    587,    640,    699,    763,    832,    908,    991,
      1081,   1180,   1288,   1405,   1534,   1673,   1826,   1993,   2175,   2373,   2590,   2826,   3084,   3365,   3672,   4008,
      4373,   4772,   5208,   5683,   6202,   6767,   7385,   8059,   8794,   9597,  10472,  11428,  12471,  13609,  14851,  16206,
     17685,  19298,  21060,  22981,  25078,  27367,  29864,  32589, -29973, -26728, -23186, -19322, -15105, -10503,  -5481,     -1,
         1,      1,   5481,  10503,  15105,  19322,  23186,  26728,  29973, -32589, -29864, -27367, -25078, -22981, -21060, -19298,
    -17685, -16206, -14851, -13609, -12471, -11428, -10472,  -9597,  -8794,  -8059,  -7385,  -6767,  -6202,  -5683,  -5208,  -4772,
     -4373,  -4008,  -3672,  -3365,  -3084,  -2826,  -2590,  -2373,  -2175,  -1993,  -1826,  -1673,  -1534,  -1405,  -1288,  -1180,
     -1081,   -991,   -908,   -832,   -763,   -699,   -640,   -587,   -538,   -493,   -452,   -414,   -379,   -348,   -318,   -292,
      -267,   -245,   -225,   -206,   -189,   -173,   -158,   -145,   -133,   -122,   -112,   -102,    -94,    -86,    -79,    -72,
       -66,    -61,    -56,    -51,    -47,    -43,    -42,    -41,    -40,    -39,    -38,    -37,    -36,    -35,    -34,    -33,
       -32,    -31,    -30,    -29,    -28,    -27,    -26,    -25,    -24,    -23,    -22,    -21,    -20,    -19,    -18,    -17,
       -16,    -15,    -14,    -13,    -12,    -11,    -10,     -9,     -8,     -7,     -6,     -5,     -4,     -3,     -2,     -1
};

/* screen buffers */
Uint32 *current_frame = NULL, *old_frame = NULL;

/* debugging file */
FILE * debug = NULL;

void MVEPlayerAudioCB(void *userdata, Uint8 *stream, Sint32 len);
void MVEPlayerDecodeAudio(buffer * input);
void MVEPlayerDecodeVideo(SDL_Surface * screen, buffer * video_data, buffer * decoding_map, Uint32 *palette);

/********************************/
/* primary function entry point */
/********************************/

int MVEPlayer(const char * filename)
{	
	/*************************/
	/* variable declarations */
	/*************************/

	/* MVE validation tools */
	const char idstring[] = "Interplay MVE File\x1A\0\x1A\0\0\x1\x33\x11";
	const Uint8 idlen = 26;
	char fileTestString[26];

	/* timer vars */
	Uint32 timer_rate = 0, start_time = 0, current_time = 0;
	Uint16 timer_subdivision = 0;
	float ms_per_frame = 0;
	Uint8 frame_timing_adjust = 0;

	/* audio variables */
	SDL_AudioSpec * desired = NULL;
	Uint16 file_audio_flags = 0;
	buffer audio_buffer, audio_data_read, temp_audio_buffer;

	/* video variables */
	Uint16 width_blocks = 0, height_blocks = 0, palette_start = 0;
	SDL_Surface * screen = NULL;
	Uint32 * pixels = NULL, numPalColors = 0, *pal = NULL;
	Uint8 r = 0, g = 0, b = 0;
	buffer map, video;

	/* event vars for determining pause or quit from user */
	SDL_Event event;
	Uint8 keyin = 0;

	/* file handle */
	SDL_RWops * mve = NULL;

	/* data structures to handle file reads */
	chunk ch;
	opcode op;

	/* some counting vars */
	Uint16 i;

	/* initialize audio buffers */
	audio_buffer.data = NULL;
	audio_buffer.length = 0;
	audio_data_read.data = NULL;
	audio_data_read.length = 0;
	temp_audio_buffer.data = NULL;
	temp_audio_buffer.length = 0;

	/*initialize decoding map buffer */
	map.data = NULL;
	map.length = 0;

	/*initialize video data buffer */
	video.data = NULL;
	video.length = 0;

	/* init debug file */
	
	debug = fopen("C:\\Documents and Settings\\CCity\\Desktop\\mvedebug.txt", "w");

	/*****************************/
	/* end variable declarations */
	/*****************************/


	/******************************************/
	/* validate MVE, init SDL, and begin read */
	/******************************************/
	
	/* initialize file access structure */
	mve = SDL_RWFromFile(filename, "rb");

	/* confirm MVE file open */
	if(!mve) 
		return UNABLE_TO_OPEN_FILE;

	/* read the first 26 bytes */
	if(SDL_RWread(mve, fileTestString, sizeof(char), idlen) < 0) 
		return FILE_TOO_SHORT;

	/* validate MVE idstring */
	if(SDL_memcmp(fileTestString, idstring, idlen) != 0) 
		return NOT_MVE_FILE;
	
	/* open first chunk */
	if(!SDL_RWread(mve, &ch, sizeof(ch), 1)) 
		return CHUNK_READ_FAILED;

	/* read opcode (initial read) */
	if(!SDL_RWread(mve, &op, sizeof(op), 1)) 
		return OPCODE_READ_FAILED;

	/* are opcode parameters within known bounds? */
	if(op.type > 0x15 || op.version > 3) 
		return MVE_CORRUPT;

	/* initialize SDL */
	if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO|SDL_INIT_TIMER) < 0)
	{
		fprintf(stderr, "Could not initialize SDL: %s", SDL_GetError());
		return SDL_INIT_FAILURE;
	}
	atexit(SDL_Quit);

	/***************************************************/
	/* end MVE validation, SDL init, and initial reads */
	/***************************************************/
	
	/*************/
	/* main loop */
	/*************/
	while(!CHKFLAG(QUITTING))
	{
		/********************************************/
		/* event polling for quit / pause from user */
		/********************************************/
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
			case SDL_KEYDOWN:
				keyin = event.key.keysym.sym;
				break;
			case SDL_KEYUP:
				if(event.key.keysym.sym == keyin)
				{
					switch(event.key.keysym.sym)
					{
					case SDLK_q:
					case SDLK_ESCAPE:
					case SDLK_RETURN:
						SETFLAG(QUITTING);
						break;
					case SDLK_SPACE:
						if(CHKFLAG(PAUSED))
							NEGFLAG(PAUSED);
						else
							SETFLAG(PAUSED);
						if(CHKFLAG(PAUSED))
						{
							if(CHKFLAG(AUDIO_PLAYING)) 
								SDL_PauseAudio(1);
							if(CHKFLAG(TIMER_CREATED)) 
								SETFLAG(TIMER_INIT);
						}
						else if(CHKFLAG(AUDIO_PLAYING)) 
							SDL_PauseAudio(0);						
						break;
					default:
						break;
					}
					keyin = 0;
				}
				break;
			case SDL_QUIT:
				SETFLAG(QUITTING);
				break;
			default:
				break;
			}
		}

		if(CHKFLAG(PAUSED) || CHKFLAG(QUITTING)) 
			continue;

		if(CHKFLAG(TIMER_INIT))
		{
			start_time = SDL_GetTicks();
			NEGFLAG(TIMER_INIT);
		}

		/********************/
		/* opcode handling */
		/********************/
    	switch(op.type)
		{
		case STOP_PLAYBACK:

			SETFLAG(QUITTING);

			break;

		case FETCH_NEXT_CHUNK:

			if(!SDL_RWread(mve, &ch, sizeof(ch), 1)) 
				return CHUNK_READ_FAILED;

			break;

		case CREATE_TIMER:

			/* get timer rate (uint32) */
			timer_rate = SDL_ReadLE32(mve);

			/* get timer subdivision (uint16) */
			timer_subdivision = SDL_ReadLE16(mve);
			ms_per_frame = (float)(timer_rate * timer_subdivision / 1000.0);
			SETFLAG(TIMER_CREATED);
			SETFLAG(TIMER_INIT);

			break;

		case INIT_AUDIO_BUFFERS:

			/* init sdl audio settings */
			desired = malloc(sizeof(SDL_AudioSpec));

			/* strip the unknown word out */
			SDL_ReadLE16(mve);

			/* get the file audio flags */
			file_audio_flags = SDL_ReadLE16(mve);

			/* get the sample rate (frequency) */
			desired->freq = SDL_ReadLE16(mve);

			/* fill the known audiospec values */
			desired->samples = 4096;
			desired->callback = MVEPlayerAudioCB;
			desired->userdata = &audio_buffer; /* may have to change that */

			/* fill values gotten from the file audio flags */
			desired->channels = (file_audio_flags & 0x01 ? 2 : 1);
			desired->format = (file_audio_flags & 0x02 ? AUDIO_S16 : AUDIO_U8);
			
			if(desired->channels == 2)
				SETFLAG(AUDIO_STEREO);
			else
				NEGFLAG(AUDIO_STEREO);

			if(desired->format == AUDIO_S16)
				SETFLAG(SAMPLESIZE16);
			else
				NEGFLAG(SAMPLESIZE16);

			/* initialize the audio mode */
			if(SDL_OpenAudio(desired, NULL) < 0)
			{
				fprintf(stderr, "Couldn't open audio: %s", SDL_GetError());
				return COULD_NOT_OPEN_AUDIO;
			}				

			/* set the compressed flag if necessary */
			if(op.version == 1)
				if(file_audio_flags & 0x04 ? 1 : 0)
					SETFLAG(AUDIO_COMPRESSED);
			
			/* we're going to abort if is compressed and (not stereo or not AUDIO_S16) because we don't know what to do */
			/* besides, we've never seen such an MVE before... */
			if(CHKFLAG(AUDIO_COMPRESSED) && (!CHKFLAG(AUDIO_STEREO) || !CHKFLAG(SAMPLESIZE16)))
				return CANNOT_PROCESS_AUDIO;

			/* get rid of rest of opcode data */
			if(op.version == 1)
				SDL_ReadLE32(mve);
			else
				SDL_ReadLE16(mve);

			break;

		case START_AUDIO:

			SETFLAG(AUDIO_PLAYING);
			if(SDL_GetAudioStatus() != SDL_AUDIO_PLAYING)
			{
				SDL_PauseAudio(0);
			}
			break;

		case INIT_VIDEO_BUFFERS:

			/* get the requested dimensions 
			note: this is in 8x8 pixel chunks so multiply by 8 to get requested screen pixels*/
			width_blocks = SDL_ReadLE16(mve);
			height_blocks = SDL_ReadLE16(mve);

			/* ditch the rest of the opcode data (note: this data is where we can get our truecolor flag (16-bit color)*/
			if(op.version > 0)
			{
				SDL_ReadLE16(mve);
				if(op.version > 1)
					SDL_ReadLE16(mve);
			}

			/* if the dimensions are different than the current, set the new ones */
			if(SDL_GetVideoInfo()->current_w != (width_blocks << 3) || SDL_GetVideoInfo()->current_h != (height_blocks << 3))
			{
				/* set SDL_NOFRAME to avoid audio / video desynch during playback due to dragging the screen around */
				screen = SDL_SetVideoMode(width_blocks << 3, height_blocks << 3, 32, SDL_SWSURFACE|SDL_ANYFORMAT);

				/* erase old video buffer data */
				if(old_frame)
				{
					free(old_frame);
					old_frame = NULL;
				}
				if(current_frame)
				{
					free(current_frame);
					current_frame = NULL;
				}

				/* allocate memory for the backbuffers sufficient for the screen pixel buffer */
				old_frame = malloc(screen->h * screen->w * sizeof(Uint32));
				assert(old_frame);
				current_frame = malloc(screen->h * screen->w * sizeof(Uint32));
				assert(current_frame);
			}

			break;

		case SEND_BUFFER_TO_DISPLAY:

			/* if there's time to next scheduled update, force delay to ms_per_frame */
			current_time = SDL_GetTicks();
			if(current_time < start_time + (Uint32)ms_per_frame)
				SDL_Delay(start_time + (Uint32)ms_per_frame - current_time);

			/* send the buffer to the screen */
			SDL_UpdateRect(screen, 0, 0, 0, 0);

			/* strip opcode data; don't know how to catch errors here */
			mve->seek(mve, op.length, SEEK_CUR);
			
			/* increment start time */
			/* note: this fix is for 66+2/3 ms_per_frame; need a general algorithm for figuring this out.  
			   frame_timing_adjust adds 1 to the ms_per_frame n - 1 out of n frames, where n is the right hand side of 
			   frame_timing_adjust %= n ; to get it to do the inverse, remove one ! from !(!frame_timing_adjust) */
			frame_timing_adjust++;
			frame_timing_adjust %= 3;
			start_time += (Uint32)ms_per_frame + !(!frame_timing_adjust);

			break;

		case AUDIO_FRAME:

			/* strip the seq-index and stream-mask */
			SDL_ReadLE16(mve);
			SDL_ReadLE16(mve);

			/* lock SDL out of the audio_buffer; we'll release at the end */
			SDL_LockAudio();

			if(CHKFLAG(AUDIO_COMPRESSED))
			{
				/* send in the raw data via audio_data_read; it will be returned the same way */
				/* read in the compressed data (should be op.length - 4 [the 4 being seq-index and stream-mask]) */
				audio_data_read.length =  op.length - 4;
				audio_data_read.data = malloc(audio_data_read.length);
				assert(audio_data_read.data != NULL);

				if(SDL_RWread(mve, audio_data_read.data, audio_data_read.length, 1) < 0)
					return AUDIO_FRAME_READ_FAILURE;
				MVEPlayerDecodeAudio(&audio_data_read);
			}
			else
			{
				/* get the stream length */
				audio_data_read.length = SDL_ReadLE16(mve);

				/* allocate memory sufficient for the data read */
				audio_data_read.data = malloc(audio_data_read.length);
				assert(audio_data_read.data != NULL);

				/* get audio data */
				if(SDL_RWread(mve, audio_data_read.data, audio_data_read.length, 1) < 0)
					return AUDIO_FRAME_READ_FAILURE;
			}
			/* at this point, we have uncompressed audio in audio_data_read, along with appropriate lengths */

			/* create temp_audio_buffer the size of old audio buffer + new audio data */
			temp_audio_buffer.length = audio_data_read.length + audio_buffer.length;
			temp_audio_buffer.data = malloc(temp_audio_buffer.length);
			assert(temp_audio_buffer.data != NULL);

			/* copy old audio buffer to temp_audio_buffer */
			memcpy(temp_audio_buffer.data, audio_buffer.data, audio_buffer.length);

			/* append new data to temp_audio_buffer */
			memcpy(temp_audio_buffer.data + audio_buffer.length, audio_data_read.data, audio_data_read.length);

			/* free audio buffer */
			if(audio_buffer.data)
				free(audio_buffer.data);

			/* temp_audio_buffer has the requested data, and audio_buffer needs to have it. */
			audio_buffer.data = temp_audio_buffer.data;
			audio_buffer.length = temp_audio_buffer.length;

			/* let SDL have access to the audio_buffer */
			SDL_UnlockAudio();

			/* close off temp_audio_buffer */
			temp_audio_buffer.data = NULL;
			temp_audio_buffer.length = 0;

			/* close off audio_data_read */
			if(audio_data_read.data)
				free(audio_data_read.data);
			audio_data_read.data = NULL;
			audio_data_read.length = 0;

			break;

		case SET_PALETTE:

			/* free previous palette if applicable */
			if(pal)
				free(pal);
			
			/* get # of first entry to fill */
			palette_start = SDL_ReadLE16(mve);
			
			/* get number of entries in palette */
			numPalColors = SDL_ReadLE16(mve);

			/* allocate memory for the palette */
			pal = malloc(sizeof(Uint32) * numPalColors);
			assert(pal != NULL);

			/* fill the palette */
			for(i = palette_start - 1; i < numPalColors; i++)
			{
				SDL_RWread(mve, &r, sizeof(Uint8), 1);
				SDL_RWread(mve, &g, sizeof(Uint8), 1);
				SDL_RWread(mve, &b, sizeof(Uint8), 1);

				/* have to store it as argb */
				/* without the +2, it's too dim */
				pal[i] = (0 << 24) | (r << (16 + 2)) | (g << (8 + 2)) | b << (0 + 2);
			}

			break;

		case SET_DECODING_MAP:

			/* kill previous map */
			if(map.data)
				free(map.data);
			
			/* get new map */
			map.data = malloc(sizeof(Uint8) * op.length);
			assert(map.data != NULL);
			map.length = op.length;
			SDL_RWread(mve, map.data, op.length, 1);
			break;

		case VIDEO_DATA:

			/* free the previously allocated video_data buffer */
			if(video.data)
				free(video.data);

			/* allocate enough memory for the video_data buffer */
			video.data = malloc(op.length);
			assert(video.data != NULL);
			video.length = op.length;

			/* read the data */
			SDL_RWread(mve, video.data, op.length, 1);

			/* lock the screen buffer so we can do stuff */
			if(SDL_MUSTLOCK(screen))
				SDL_LockSurface(screen);

			/* decode the video data */
			MVEPlayerDecodeVideo(screen, &video, &map, pal);
			
			/* unlock the screen buffer so SDL can do stuff */
			if(SDL_MUSTLOCK(screen))
				SDL_UnlockSurface(screen);

			break;

		default:

			/* are opcode parameters within known bounds? */
			if(op.type > 0x15 || op.version > 3) 
				return MVE_CORRUPT;

			/* strip opcode data; don't know how to catch errors here */
			mve->seek(mve, op.length, SEEK_CUR);

			break;
		}

		/************************/
		/* end opcode handling */
		/************************/
		
		/* read opcode (trailing read) */
		if(!SDL_RWread(mve, &op, sizeof(op), 1)) 
			return OPCODE_READ_FAILED;

		/* are opcode parameters within known bounds? */
		if(op.type > 0x15 || op.version > 3) 
			return MVE_CORRUPT;

	}
	/***********/
	/* cleanup */
	/***********/

	/* stop audio */
	if(CHKFLAG(AUDIO_PLAYING))
	{
		SDL_PauseAudio(1);
		SDL_CloseAudio();
	}

	/* free audio buffers */
	if(audio_buffer.data)
		free(audio_buffer.data);
	if(audio_data_read.data)
		free(audio_data_read.data);
	if(temp_audio_buffer.data)
		free(temp_audio_buffer.data);

	/* free the audio_spec */
	if(desired)
		free(desired);

	/* free the palette */
	if(pal)
		free(pal);

	/* free the decoding map */
	if(map.data)
		free(map.data);

	/* free the video data */
	if(video.data)
		free(video.data);

	/* free the video buffers */
	if(old_frame)
		free(old_frame);

	if(current_frame)
		free(current_frame);

	/* close the file handle */
	if(mve)
		SDL_RWclose(mve);

	if(debug)
		fclose(debug);

	/* seems to have worked. */
	return SUCCESS;
}

/******************************/
/* Audio callback entry point */
/******************************/
void MVEPlayerAudioCB(void *userdata, Uint8 *stream, int len)
{
	Uint8 * temp;
	buffer *audio_buffer = (buffer *)userdata;

	if((Uint32)len > audio_buffer->length)
	{
		/* fprintf(stderr, "SDL wants more data than we have (%i, vs. %i).\n", len, audio_buffer->length); */
		/* give SDL what we have */
		memcpy(stream, audio_buffer->data, audio_buffer->length);
		/* and now we have nothing */
		audio_buffer->length = 0;
	}
	else
	{
		/* copy len bytes to stream */
		memcpy(stream, audio_buffer->data, len);

		/* our buffer length is now whatever we started with minus len */
		audio_buffer->length -= len;

		/* now we need to store the data that wasn't read */
		temp = malloc(audio_buffer->length);
		assert(temp != NULL);

		/* copy from data address + len length bytes 
		(which should be the start of the data that wasn't copied to stream) */
		memcpy(temp, audio_buffer->data + len, audio_buffer->length);

		/* free the memory of the audio_buffer */
		if(audio_buffer->data)
			free(audio_buffer->data);

		/* assign the address of our stored data to the audio_buffer's data */
		audio_buffer->data = temp;
		
		/* close off temp */
		temp = NULL;
	}
}

/*****************************/
/* audio decoder entry point */
/*****************************/
void MVEPlayerDecodeAudio(buffer * in)
{
	/* this only works for stereo dpcm at present */

	Sint16 sample[2];
	Uint16 in_pos = 0, out_pos = 0, channel = 0;
	buffer out;

	/* in->data includes initial stream-len word */
	/* uncompressed streamlen is stream-len bytes */
	out.length = LE16(in->data);
	in_pos +=2;
	out.data = malloc(out.length);
	assert(out.data != NULL);

	/* each byte in the input buffer after the first four (two words: initial L and R values)
	will be expanded to fill a word in the return buffer after decompression. 
	The initial two words will be stored in return buffer as-is. */
	sample[0] = LE16(in->data + in_pos);
	in_pos += 2;
	WRITEWORD((out.data + out_pos), sample[0]);
	out_pos += 2;
	sample[1] = LE16(in->data + in_pos);
	in_pos += 2;
	WRITEWORD((out.data + out_pos), sample[1]);
	out_pos += 2;

	while(in_pos < in->length)
	{
		sample[channel] += interplay_delta_table[*(in->data + in_pos++)];
		WRITEWORD((out.data + out_pos), sample[channel]);
		out_pos += 2;
		channel++;
		channel %= 2;
	}

	/* free the original buffer*/
	if(in->data)
		free(in->data);

	/* assign the incoming buffer the outgoing data for return */
	in->data = out.data;
	in->length = out.length;
}

/**************************************/
/* Video decoding routine entry point */
/**************************************/
void MVEPlayerDecodeVideo(SDL_Surface * screen, buffer * pData, buffer * pMap, Uint32 * pal)
{
#define CHECKBUF(exp) {if(data.length < exp)\
	{fprintf(debug,"\nout of video data (%i requested, %i present), encoding: %X, block (%i, %i)\n",\
	exp, data.length, encoding, x, y); exit(12);}else data_tracker += exp;}

#define INCBUF(exp) {exp.data++;exp.length--;}

  /*Opcode 0x11: Video Data

    Ok, this is the big killer opcode.  The way this works is as follows:

    First, the data is processed in 8x8 pixel blocks.  There are 4 bits
    associated with each block giving the particular encoding to use for that
    block, giving a total of 16 possible encodings for a given block.  These
    4-bit pieces come from the most recent 0xf opcode data stream.  They all
    appear to be used (or at least supported by the player).  So, I'll go over
    the encodings for each of the 16 encoding types.  The rendering process
    keeps track of the most recent frame in a separate buffer, and uses this
    double-buffering technique in the common way for animation.  The current
    frame's data is used in the construction of the next frame.  In the
    following description, "current frame" will refer to the most recently
    displayed frame, and "new frame" will refer to the frame currently being
    constructed for display.  "map stream" will refer to the data grabbed from
    the 0xf Opcode data, and "data stream" will refer to the data grabbed from
    the 0x11 opcode data.*/

	Uint8 hblocks = 0, wblocks = 0, encoding = 0;
	Uint32 screen_buffer_size, *new_frame = NULL, *current = NULL, *old = NULL, data_tracker = 0;
	Sint32 x, y, i, j;
	buffer map, data;

	map.data = pMap->data;
	map.length = pMap->length;

	/* there are 14 undocumented bytes at head of video stream */
	data.data = pData->data + 14;
	data.length = pData->length - 14;

	screen_buffer_size = screen->h * screen->w * sizeof(Uint32);

	/* just making sure they were initialized */
	assert(old_frame);
	assert(current_frame);
	
	/* copy current_frame into old_frame */
	memcpy(old_frame, current_frame, screen_buffer_size);

	/* copy the most recently displayed frame into current_frame */
	memcpy(current_frame, screen->pixels, screen_buffer_size);

	/* copy the old_frame into the new frame */
	memcpy(screen->pixels, old_frame, screen_buffer_size);

	/* read our number of blocks by dividing width and height by 8 */
	hblocks = screen->h >> 3;
	wblocks = screen->w >> 3;

	// for each row (of 8x8 pixel blocks)
    for(y = 0; y < hblocks; y++) 
	{
		// for each column (of 8x8 pixel blocks)
		for (x = 0; x < wblocks; x++) 
		{
			/* assign the frame buffer pointers to the addresses of the screen 
			buffers so we can do pointer arithmetic on them such as incrementing. */

			new_frame = screen->pixels;
			old = old_frame;
			current = current_frame;

			/* increment pixel pointer past the rendered area on the screen
			y = number of rows, wblocks = number of blocks per row, 8 * 8 = pixel area of the block; 
			y * wblocks * 8 * 8 = all the complete rows of 8x8 pixel blocks rendered.
			x * 8 = upper left pixel of the 8x8 block in this new row to be rendered at this time. */

			new_frame += (y * wblocks * 8 * 8) + (x * 8);
			old += (y * wblocks * 8 * 8) + (x * 8);
			current += (y * wblocks * 8 * 8) + (x * 8);
				
			/* get the decoding map info for this block; low 4 bits if x is even, high 4 bits if x is odd */
			encoding = (*map.data >> (4 * (x & 1))) & 0x0f;

			/* if x is odd, increment the decoding map, and decrement length */
			if(x % 2)
				INCBUF(map);

			/* process the block */
			switch(encoding)
			{
			case 0x0:

				/* Encoding 0x0:
            Block is copied from corresponding block from current frame.
            (i.e. this block is unchanged). */

				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
					{
						new_frame[i] = current[i];
					}
					/* wrap around to next row */
					new_frame += screen->w;
					current += screen->w;
				}
				break;

			case 0x1:

				/* Encoding 0x1:
            Block is unmodified.  This appears to mean that it has the same
            value it had 2 frames ago, but the net effect is that nothing is
            done to this block of 8x8 pixels. */

				break;

			case 0x2:

				/* Encoding 0x2:
            Block is copied from nearby (below and to the right) within the
            new frame.  The offset within the buffer from which to grab the
            patch of 8 pixels is given by grabbing a byte B from the data
            stream, which is broken into a positive x and y offset according
            to the following mapping:

                if B < 56:
                    x = 8 + (B % 7)
                    y = B / 7
                else
                    x = -14 + ((B - 56) % 29)
                    y =   8 + ((B - 56) / 29)

            (where % is the 'modulo' operator)

            If you draw the region this represents, you'll see it looks like:

                       oooooooo#######
                       oooooooo#######
                       oooooooo#######
                       oooooooo#######
                       oooooooo#######
                       oooooooo#######
                       oooooooo#######
                       oooooooo#######
         #############################
         #############################
         #############################
         #############################
         #############################
         #############################
         ##########################

         Where 'o' are the pixels in the destination frame, and # are the
         locations where the source frame could start. */
				{
					Uint32 *pixels = new_frame;
					if(*data.data < 56)
					{
						i = 8 + (*data.data % 7);
						j = *data.data / 7;
					}
					else
					{
						i = -14 + ((*data.data - 56) % 29);
						j =   8 + ((*data.data - 56) / 29);
					}
					INCBUF(data);

					new_frame += j * screen->w + i;

					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
							pixels[i] = new_frame[i];
						}
						pixels += screen->w;
						new_frame += screen->w;
					}
				}

				break;

			case 0x3:

				/* Encoding 0x3:
            This is the same as encoding 0x2, with the exception that the x
            and y offsets are negated giving:

                if B < 56:
                    x = -(8 + (B % 7))
                    y = -(B / 7)
                else
                    x = -(-14 + ((B - 56) % 29))
                    y = -(  8 + ((B - 56) / 29))

            (where % is the 'modulo' operator)

            If you draw the region this represents, you'll see it looks like:

                   ##########################
                #############################
                #############################
                #############################
                #############################
                #############################
                #############################
                #######oooooooo
                #######oooooooo
                #######oooooooo
                #######oooooooo
                #######oooooooo
                #######oooooooo
                #######oooooooo
                #######oooooooo */
				{
					Uint32 * pixels = new_frame;
					if(*data.data < 56)
					{
						i = -(8 + (*data.data % 7));
						j = -(*data.data / 7);
					}
					else
					{
						i = -(-14 + ((*data.data - 56) % 29));
						j = -(  8 + ((*data.data - 56) / 29));
					}
					INCBUF(data);

					new_frame += j * screen->w + i;

					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
							pixels[i] = new_frame[i];
						}
						pixels += screen->w;
						new_frame += screen->w;
					}
				}
				break;

			case 0x4:

				/* Encoding 0x4:
            Similar to 0x2 and 0x3, except this method copies from the
            "current" frame, rather than the "new" frame, and instead of the
            lopsided mapping they use, this one uses one which is symmetric
            and centered around the top-left corner of the block.  This uses
            only 1 byte still, though, so the range is decreased, since we
            have to encode all directions in a single byte.  The byte we pull
            from the data stream, I'll call B.  Call the highest 4 bits of B
            BH and the lowest 4 bytes BL.  Then the offset from which to copy
            the data is:

                x = -8 + BL
                y = -8 + BH */

				i = -8 + (0x0F & *data.data);
				j = -8 + ((0xF0 & *data.data) >> 4);
				INCBUF(data);

				current += j * screen->w + i;

				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
					{
						new_frame[i] = current[i];
					}
					current += screen->w;
					new_frame += screen->w;
				}
				break;

			case 0x5:

				/* Encoding 0x5:
            Similar to 0x4, but instead of one byte for the offset, this uses
            two bytes to encode a larger range, the first being the x offset
            as a signed 8-bit value, and the second being the y offset as a
            signed 8-bit value. */
				{
					Sint8 m, n;

					m = data.data[0];
					n = data.data[1];
					INCBUF(data);
					INCBUF(data);

					current += n * screen->w + m;

					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
							new_frame[i] = current[i];
						}
						current += screen->w;
						new_frame += screen->w;
					}
				}
				break;

			case 0x6:

				/* Encoding 0x6:
            I can't figure out how any file containing a block of this type
            could still be playable, since it appears that it would leave the
            internal bookkeeping in an inconsistent state in the BG player
            code.  Ahh, well.  Perhaps it was a bug in the BG player code that
            just didn't happen to be exposed by any of the included movies.
            Anyway, this skips the next two blocks, doing nothing to them.
            Note that if you've reached the end of a row, this means going on
            to the next row. */

				if(y == hblocks - 1 && x > wblocks - 3)
				{
					return;
				}

				if(x > wblocks - 3)
				{
					x =+ 2;
					/* set it into the appropriate block in the next row */
					x %= (wblocks);
					y++;
					break;
				}
				x += 2;
				break;

			case 0x7:

				/* Encoding 0x7:
            Ok, here's where it starts to get really...interesting.  This is,
            incidentally, the part where they started using self-modifying
            code.  So, most of the following encodings are "patterned" blocks,
            where we are given a number of pixel values and then bitmapped
            values to specify which pixel values belong to which squares.  For
            this encoding, we are given the following in the data stream:

                P0 P1

            These are pixel values (i.e. 8-bit indices into the palette).  If
            P0 <= P1, we then get 8 more bytes from the data stream, one for
            each row in the block:

                B0 B1 B2 B3 B4 B5 B6 B7

            For each row, the rightmost pixel is represented by the low-order
            bit, and the leftmost by the high-order bit.  Use your imagination
            in between.  If a bit is set, the pixel value is P1 and if it is
            unset, the pixel value is P0.

            If, on the other hand, P0 > P1, we get two more bytes from the
            data stream:

                B0 B1

            Each of these bytes contains a 4-bit pattern.  This pattern works
            exactly like the pattern above with 8 bytes, except each bit
            represents a 2x2 pixel region.

            So, for example, if we had:

                11 22 ff 81 81 81 81 81 81 ff

            This would represent the following layout:

                22 22 22 22 22 22 22 22     ; ff == 11111111
                22 11 11 11 11 11 11 22     ; 81 == 10000001
                22 11 11 11 11 11 11 22     ; ..
                22 11 11 11 11 11 11 22
                22 11 11 11 11 11 11 22
                22 11 11 11 11 11 11 22
                22 11 11 11 11 11 11 22     ; 81 == 10000001
                22 22 22 22 22 22 22 22     ; ff == 11111111

            If, on the other hand, we had:

                22 11 ff 81

            The output would be:

                22 22 22 22 22 22 22 22     ; f == 1 1 1 1
                22 22 22 22 22 22 22 22     ; 
                22 22 22 22 22 22 22 22     ; f == 1 1 1 1
                22 22 22 22 22 22 22 22     ; 
                22 11 11 11 11 11 11 11     ; 8 == 1 0 0 0
                22 11 11 11 11 11 11 11     ; 
                11 11 11 11 11 11 11 22     ; 1 == 0 0 0 1
                11 11 11 11 11 11 11 22     ;  */
				{
					Uint8 p[2], b[8], mask;
					p[0] = data.data[0];
					p[1] = data.data[1];
					INCBUF(data);
					INCBUF(data);
					if(p[0] > p[1])
					{
						b[0] = data.data[0];
						b[1] = data.data[1];
						INCBUF(data);
						INCBUF(data);
						for(j = 0; j < 4; j++)
						{
							/* if j is even (0 or 2) */
							if(!(j & 1))
							{
								/* set the mask bit (10000000b) */
								mask = 0x80;
							}
							for(i = 0; i < 4; i++)
							{
								/* color the whole 2x2 block */
								new_frame[i * 2] = 
									new_frame[i * 2 + 1] = 
									new_frame[screen->w + i * 2] =
									new_frame[screen->w + i * 2 + 1] = 
									pal[p[!(mask & b[j > 1 ? 1: 0])]];
								mask >>= 1;
							}
							/* should put us 2 rows below the corner we started on */
							new_frame += screen->w * 2;
						}
					}
					else
					{
						for(j = 0; j < 8; j++)
						{
							b[j] = data.data[0];
							INCBUF(data);
						}
						
						for(j = 0; j < 8; j++)
						{
							mask = 0x80;
							for(i = 0; i < 8; i++)
							{
								new_frame[i] = pal[p[!(mask & b[j])]];
								mask >>= 1;
							}
							new_frame += screen->w;
						}
					} /* end if / else block */
				} /* end case block */
				break;

			case 0x8:

				/* Encoding 0x8:
            Ok, this one is basically like encoding 0x7, only more
            complicated.  Again, we start out by getting two bytes on the data
            stream:

                P0 P1

            if P0 <= P1 then we get the following from the data stream:

                      B0 B1
                P2 P3 B2 B3
                P4 P5 B4 B5
                P6 P7 B6 B7

            P0 P1 and B0 B1 are used for the top-left corner, P2 P3 B2 B3 for
            the bottom-left corner, P4 P5 B4 B5 for the top-right, P6 P7 B6 B7
            for the bottom-right.  (So, each codes for a 4x4 pixel array.)
            Since we have 16 bits in B0 B1, there is one bit for each pixel in
            the array.  The convention for the bit-mapping is, again, left to
            right and top to bottom.

            So, basically, the top-left quarter of the block is an arbitrary
            pattern with 2 pixels, the bottom-left a different arbitrary
            pattern with 2 different pixels, and so on.  I'll go through a few
            examples of this after I discuss the other forms for the data in
            this encoding.

            if P0 > P1 then we get 10 more bytes from the data stream:

                B0 B1 B2 B3 P2 P3 B4 B5 B6 B7

            Now, if P2 <= P3, then [P0 P1 B0 B1 B2 B3] represent the left half
            of the block and [P2 P3 B4 B5 B6 B7] represent the right half.

            If P2 > P3, [P0 P1 B0 B1 B2 B3] represent the top half of the
            block and [P2 P3 B4 B5 B6 B7] represent the bottom half.

            In these last two cases, each bit represents a 1x1 pixel.  Just to
            work through an example of each case:

                00 22 f9 9f 11 33 cc 33 44 55 aa 55 66 77 01 ef

                22 22 22 22 | 33 33 11 11     ; f = 1111, c = 1100
                22 00 00 22 | 33 33 11 11     ; 9 = 1001, c = 1100
                22 00 00 22 | 11 11 33 33     ; 9 = 1001, 3 = 0011
                22 22 22 22 | 11 11 33 33     ; f = 1111, 3 = 0011
                ------------+------------
                55 44 55 44 | 66 66 66 66     ; a = 1010, 0 = 0000
                55 44 55 44 | 66 66 66 77     ; a = 1010, 1 = 0001
                44 55 44 55 | 77 77 77 66     ; 5 = 0101, e = 1110
                44 55 44 55 | 77 77 77 77     ; 5 = 0101, f = 1111

            I've added a dividing line in the above to clearly delineate the
            quadrants.

            Now, for a horizontally split block:

                22 00 01 37 f7 31 11 66 8c e6 73 31

                22 22 22 22 66 11 11 11
                22 22 22 00 66 66 11 11
                22 22 00 00 66 66 66 11
                22 00 00 00 11 66 66 11
                00 00 00 00 11 66 66 66
                22 00 00 00 11 11 66 66
                22 22 00 00 11 11 66 66
                22 22 22 00 11 11 11 66

            Finally, for a vertically split block:

                22 00 cc 66 33 19 66 11 18 24 42 81

                00 00 22 22 00 00 22 22
                22 00 00 22 22 00 00 22
                22 22 00 00 22 22 00 00
                22 22 22 00 00 22 22 00
                66 66 66 11 11 66 66 66
                66 66 11 66 66 11 66 66
                66 11 66 66 66 66 11 66
                11 66 66 66 66 66 66 11 */
				{
					Uint8 p[8], b[8];
					p[0] = data.data[0];
					p[1] = data.data[1];
					INCBUF(data);
					INCBUF(data);
					if(p[0] <= p[1])
					{
						b[0] = data.data[0];
						b[1] = data.data[1];
						INCBUF(data);
						INCBUF(data);
						for(i = 1; i < 4; i++)
						{
							p[i * 2] = data.data[0];
							p[i * 2 + 1] = data.data[1];
							b[i * 2] = data.data[2];
							b[i * 2 + 1] = data.data[3];
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
						}
					}
					else
					{
						for(i = 0; i < 4; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}

						p[2] = data.data[0];
						p[3] = data.data[1];
						INCBUF(data);
						INCBUF(data);

						for(i = 4; i < 8; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
					} /* end if / else block */
				} /* end case block */

				break;

			case 0x9:

				/* Encoding 0x9:
            Similar to the previous 2 encodings, only more complicated.  And
            it will get worse before it gets better.  No longer are we dealing
            with patterns over two pixel values.  Now we are dealing with
            patterns over 4 pixel values with 2 bits assigned to each pixel
            (or block of pixels).

            So, first on the data stream are our 4 pixel values:

                P0 P1 P2 P3

            Now, if P0 <= P1  AND  P2 <= P3, we get 16 bytes of pattern, each
            2 bits representing a 1x1 pixel (00=P0, 01=P1, 10=P2, 11=P3).  The
            ordering is again left to right and top to bottom.  The most
            significant bits represent the left side at the top, and so on.

            If P0 <= P1  AND  P2 > P3, we get 4 bytes of pattern, each 2 bits
            representing a 2x2 pixel.  Ordering is left to right and top to
            bottom.

            if P0 > P1  AND  P2 <= P3, we get 8 bytes of pattern, each 2 bits
            representing a 2x1 pixel (i.e. 2 pixels wide, and 1 high).

            if P0 > P1  AND  P2 > P3, we get 8 bytes of pattern, each 2 bits
            representing a 1x2 pixel (i.e. 1 pixel wide, and 2 high). */
				{
					Uint8 p[4], b[16];
					for(i = 0; i < 4; i++)
					{
						p[i] = data.data[0];
						INCBUF(data);
					}
					if(p[0] <= p[1] && p[2] <= p[3])
						for(i = 0; i < 16; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
					if(p[0] <= p[1] && p[2] > p[3])
						for(i = 0; i < 4; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
					if(p[0] > p[1] && p[2] <= p[3])
						for(i = 0; i < 8; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
					if(p[0] > p[1] && p[2] > p[3])
						for(i = 0; i < 8; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
				}
				break;

			case 0xA:

				/* Encoding 0xa:
            Similar to the previous, only a little more complicated.
            We are still dealing with patterns over 4 pixel values with 2 bits
            assigned to each pixel (or block of pixels).

            So, first on the data stream are our 4 pixel values:

                P0 P1 P2 P3

            Now, if P0 <= P1, the block is divided into 4 quadrants, ordered
            (as with opcode 0x8) TL, BL, TR, BR.  In this case the next data
            in the data stream should be:

                                B0  B1  B2  B3
                P4  P5  P6  P7  B4  B5  B6  B7
                P8  P9  P10 P11 B8  B9  B10 B11
                P12 P13 P14 P15 B12 B13 B14 B15

            Each 2 bits represent a 1x1 pixel (00=P0, 01=P1, 10=P2, 11=P3).
            The ordering is again left to right and top to bottom.  The most
            significant bits represent the left side at the top, and so on.

            If P0 > P1 then the next data on the data stream is:

                            B0 B1 B2  B3  B4  B5  B6  B7
                P4 P5 P6 P7 B8 B9 B10 B11 B12 B13 B14 B15

            Now, in this case, if P4 <= P5,
            [P0 P1 P2 P3 B0 B1 B2 B3 B4 B5 B6 B7] represent the left half of
            the block and the other bytes represent the right half.  If P4 >
            P5, then [P0 P1 P2 P3 B0 B1 B2 B3 B4 B5 B6 B7] represent the top
            half of the block and the other bytes represent the bottom half. */
				{
					Uint8 p[16], b[16];
					for(i = 0; i < 4; i++)
					{
						p[i] = data.data[0];
						INCBUF(data);
					}
					if(p[0] <= p[1])
					{
						for(i = 0; i < 4; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
						for(i = 1; i < 4; i++)
						{
							p[i * 4] = data.data[0];
							p[i * 4 + 1] = data.data[1];
							p[i * 4 + 2] = data.data[2];
							p[i * 4 + 3] = data.data[3];
							b[i * 4] = data.data[4];
							b[i * 4 + 1] = data.data[5];
							b[i * 4 + 2] = data.data[6];
							b[i * 4 + 3] = data.data[7];
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
							INCBUF(data);
						}
					}
					if(p[0] > p[1])
					{
						for(i = 0; i < 8; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
						for(i = 4; i < 8; i++)
						{
							p[i] = data.data[0];
							INCBUF(data);
						}
						for(i = 8; i < 16; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
					}
				}
				break;

			case 0xB:

				/* Encoding 0xb:
            In this encoding we get raw pixel data in the data stream -- 64
            bytes of pixel data.  1 byte for each pixel, and in the standard
            order (l->r, t->b). */
				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
					{
						new_frame[i] = pal[data.data[0]];
						INCBUF(data);
					}
					new_frame += screen->w;
				}
				break;

			case 0xC:

				/* Encoding 0xc:
            In this encoding we get raw pixel data in the data stream -- 16
            bytes of pixel data.  1 byte for each block of 2x2 pixels, and in
            the standard order (l->r, t->b). */
				{
					Uint8 p[16];
					for(i = 0; i < 16; i++)
					{
						p[i] = data.data[0];
						INCBUF(data);
					}
				}
				break;

			case 0xD:

				/* Encoding 0xd:
            In this encoding we get raw pixel data in the data stream -- 4
            bytes of pixel data.  1 byte for each block of 4x4 pixels, and in
            the standard order (l->r, t->b). */
				{
					Uint8 p[4];
					for(i = 0; i < 4; i++)
					{
						p[i] = data.data[0];
						INCBUF(data);
					}
				}
				break;

			case 0xE:

				/* Encoding 0xe:
            This encoding represents a solid frame.  We get 1 byte of pixel
            data from the data stream. */
				for(j = 0; j < 8; j++) 
				{
					for(i = 0; i < 8; i++)
						new_frame[j] = pal[data.data[0]];
					new_frame += screen->w;
				}
				INCBUF(data);
				break;

			case 0xF:

				/* Encoding 0xf:
            This encoding represents a "dithered" frame, which is
            checkerboarded with alternate pixels of two colors.  We get 2
            bytes of pixel data from the data stream, and these bytes are
            alternated:

                P0 P1 P0 P1 P0 P1 P0 P1
                P1 P0 P1 P0 P1 P0 P1 P0
                ...
                P0 P1 P0 P1 P0 P1 P0 P1
                P1 P0 P1 P0 P1 P0 P1 P0 */

				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
						new_frame[i] = pal[data.data[(i + j) & 1]];
					new_frame += screen->w;
				}
				INCBUF(data);
				INCBUF(data);
				break;
			} /* end switch(encoding) */
		} /* end for(x) */
	} /* end for(y) */
}